#include <string>
#include <sstream>
#include <iostream>
#include <fstream>
#include "Tree.hpp"
#include "Node.hpp"

namespace nbody {
	using namespace std;

	Tree::Tree(int parallelId) {
		//insert dummy root node
		this->nodes = new Node(this);
		this->maxLeafBodies = 16;
		this->parallelId = parallelId;
	}

	Tree::~Tree() {
		this->clean();
		delete this->nodes;
	}

	void Tree::clean() {
		//remove all nodes; refresh dummy first node
		while (this->nodes->next != this->nodes) {
			Node* node = this->nodes->next;

			node->unlink();
			delete node;
		}
		delete this->nodes;
		this->nodes = new Node(this);
	}

	unsigned long Tree::numberOfNodes() {
		unsigned long nodes = 0;

		for (Node* node = this->nodes->next; node != this->nodes; node = node->next) {
			nodes++;
		}
		return nodes;
	}

	bool Tree::isCorrect() {
		Node* current = this->nodes->next;

		while (current != this->nodes) {
			if (!current->isCorrect()) {
				return false;
			}
			current = current->next;
		}
		return true;
	}

	void Tree::accumulateForceOnto(Body& body) {
		Node* n = this->nodes->next;

		resetAcceleration(body);
		while (n != this->nodes) {
			//cout << "n " << n << " " << n->leaf << endl;
			if (n->sufficientForBody(body)) {
				accumulateForceOntoBody(body, n->representative);
				n = n->afterSubtree;
				//cout << "rb" << endl;
			} else if (n->leaf) {
				//cout << "l" << endl;
				for (vector<Body>::iterator it = n->bodies.begin(); it != n->bodies.end(); it++) {
					//cout << "lb" << endl;
					accumulateForceOntoBody(body, *it);
				}
				n = n->afterSubtree;
			} else {
				n = n->next;
			}
			//n = n->afterSubtree;
		}
		//body.print();
	}

	void Tree::computeForces() {
		for (Node* n = this->nodes->next; n != this->nodes; n = n->next) {
			//cout << "Body" << endl;
			if (n->leaf) {
				for (vector<Body>::iterator it = n->bodies.begin(); it != n->bodies.end(); it++) {
					this->accumulateForceOnto(*it);
				}
			}
		}
	}

	vector<Body> Tree::extractLocalBodies() {
		vector<Body> result;

		while (this->nodes->next != this->nodes) {
			if (this->nodes->next->leaf) {
				for (vector<Body>::iterator it = this->nodes->next->bodies.begin(); it != this->nodes->next->bodies.end(); it++) {
					if (!it->refinement) {
						result.push_back(*it);
					}
				}
			}
			Node* h = this->nodes->next;
			this->nodes->next->unlink();
			delete(h);
		}
		this->clean();
		return result;
	}

	vector<Body> Tree::copyRefinements(Box domain) {
		vector<Body> result;
		Node* current = this->nodes->next;

		if (!isValid(current->bb)) {
			//empty tree means no refinements
			//cout << "invalid" << endl;
			return result;
		}
		while (current != this->nodes) {
			//cout << "trav node" << endl;
			bool sufficient = current->sufficientForBox(domain);

			if (sufficient) {
				result.push_back(current->representative);
				current = current->afterSubtree;
			} else if (current->leaf) {
				result.insert(result.end(), current->bodies.begin(), current->bodies.end());
				current = current->next;
			} else {
				current = current->next;
			}
		}
		return result;
	}

	Box Tree::getRootBB() {
		return this->nodes->next->bb;
	}

	void Tree::rebuild(Box domain) {
		this->build(this->extractLocalBodies(), domain);
	}

	void Tree::rebuild() {
		this->build(this->extractLocalBodies());
	}

	void Tree::rebuild(Box domain, vector<Body> bodies) {
		this->build(bodies, domain);
	}

	Box Tree::advance() {
		Box bb;

		for (Node* n = this->nodes->next; n != this->nodes; n = n->next) {
			if (n->leaf) {
				for (vector<Body>::iterator it = n->bodies.begin(); it != n->bodies.end(); it++) {
					if (!it->refinement) {
						integrate(*it);
						//it->print();
						extend(bb, *it);
					}
				}
			}
		}
		//bb.print(-1);
		return bb;
	}

	Box Tree::getLocalBB() {
		Box result;

		for (Node* n = this->nodes->next; n != this->nodes; n = n->next) {
			if (n->leaf) {
				for (vector<Body>::iterator it = n->bodies.begin(); it != n->bodies.end(); it++) {
					if (!it->refinement) {
						extend(result, *it);
					}
				}
			}
		}
		return result;
	}

	void Tree::print(int parallelId) {
		printBB(parallelId, this->nodes->next->bb);
	}
}
